import pygame, sys,os, math, curves, data, random
from pygame.locals import *
from locals import *

def loadMusic(name):
	return pygame.mixer.Sound(data.load(name+'.ogg'))

def loadSound(name):
	return pygame.mixer.Sound(data.load(name+'.wav'))

def loadImage(filename, mode=0, colorkey=(255,0,255)):
	if mode == IMG_COLORKEY:
		surface = pygame.image.load(data.load(filename)).convert()
		surface.set_colorkey(colorkey)
		return surface
	elif mode == IMG_PNG24:
		return pygame.image.load(data.load(filename)).convert_alpha()
	else:
		return pygame.image.load(data.load(filename)).convert()

def spawnSingle(sprite, spawnPoint, spritegroup):
	spriteGroup.add(self.sprite)



'''
######################## Sprite-Classes ##########################
'''


class MousePointer (pygame.sprite.Sprite):
	def __init__(self, pg):
		pygame.sprite.Sprite.__init__(self)
		self.pg = pg
		self.image = loadImage('mouse_pointer.png', IMG_COLORKEY)
		self.rect = self.image.get_rect()
		pygame.mouse.set_pos(self.pg.pos[0]+100, 200)

	def update(self):
		self.rect.center = pygame.mouse.get_pos()
		if self.rect.x < self.pg.pos[0]+10:
			self.rect.x = self.pg.pos[0]+10

		if self.rect.x > self.pg.pos[0] + self.pg.size[0] - 16:
			self.rect.x = self.pg.pos[0] + self.pg.size[0] - 16

		if self.rect.y > self.pg.size[1] - 26:
			self.rect.y = self.pg.size[1] - 26

		if self.rect.y < 10:
			self.rect.y = 10


class Laser (pygame.sprite.Sprite):
	def __init__(self, spawnPoint, speed=800):
		pygame.sprite.Sprite.__init__(self)
		self.image = loadImage('laser2.png', IMG_PNG24)
		self.rect = self.image.get_rect()
		self.rect.center = self.x, self.y = self.spawnPoint = spawnPoint
		self.startTime = pygame.time.get_ticks()
		self.speed = speed / 1000.0 #yPerMilliseconds
		
	def update(self):
		self.calculatePosition()
		self.rect.center = self.x, self.y
		if self.rect.y < 0:
			self.kill()

	def calculatePosition(self):
		self.lifeTime = pygame.time.get_ticks() - self.startTime
		self.y = self.spawnPoint[1] - self.speed * self.lifeTime



class Player (pygame.sprite.Sprite):
	def __init__(self, spawnPoint, pg):
		pygame.sprite.Sprite.__init__(self)
		self.pg = pg
		self.base_image = loadImage('enemy2.png', IMG_COLORKEY)
		self.image = self.base_image.subsurface((0,0,22,27))
		self.image = pygame.transform.scale(self.image, (44,54))
		self.rect = self.image.get_rect()
		self.x, self.y = self.rect.center = (spawnPoint)
		self.rect.y = 500
		pygame.mouse.set_visible(False)
		self.sounds = []
		self.sounds.append(loadSound('player-shoot-01'))
		self.sounds.append(loadSound('player-shoot-02'))
		self.sounds.append(loadSound('player-shoot-03'))

		self.a = 0.02
		self.vx = self.vy = 0.0
		self.maxV = 1.5
		self.damp = 0.99

	def update(self):
		if self.pg.world.input_left:
			self.stearLeft()
		if self.pg.world.input_right:
			self.stearRight()
		if self.pg.world.input_up:
			self.stearUp()
		if self.pg.world.input_down:
			self.stearDown()

		if self.pg.world.input_mouse_button_left and self.pg.world.ready_to_fire:
			self.shootLaser()
		

		self.vx *= self.damp			
		self.vy *= self.damp

		self.x += self.vx
		self.y += self.vy


		if self.y < 200:
			self.y = 200
			self.vy = 0.0
		elif self.y > self.pg.size[1] - self.rect.height/2:
			self.y = self.pg.size[1] - self.rect.height/2
			self.vy = 0.0
		if self.x > self.pg.size[0] - self.rect.width/2:
			self.x = self.pg.size[0] - self.rect.width/2
			self.vx = 0.0
		elif self.x < 0 + self.rect.width/2:
			self.x = self.rect.width/2
			self.vx = 0.0

		self.rect.center = self.x, self.y


	def stearLeft(self):
		self.vx -= self.a
		if self.vx <= -self.maxV:
			self.vx = -self.maxV
		
	def stearRight(self):
		self.vx += self.a
		if self.vx >= self.maxV:
			self.vx = self.maxV

	def stearUp(self):
		self.vy -= self.a
		if self.vy <= -self.maxV:
			self.vy = -self.maxV
		
	def stearDown(self):
		self.vy += self.a
		if self.vy >= self.maxV:
			self.vy = self.maxV


	def loadImage(self):
		self.base_image = pygame.image.load(data.load('player.png')).convert()
		self.base_image.set_colorkey((255,0,255))
		self.image = pygame.transform.rotate(self.base_image,0)
		self.rect = self.image.get_rect()

	def shootLaser(self):
		self.pg.world.ready_to_fire = False
		sound = random.choice(self.sounds)
		sound.set_volume(0.4)
		sound.play()
		self.pg.laserGroup.add(Laser((self.rect.centerx, self.rect.y-2)))


class Enemy (pygame.sprite.Sprite):
	def __init__(self, pg, base_image):
		pygame.sprite.Sprite.__init__(self)
		self.pg = pg
		self.lifeTime = 0
		self.startTime = pygame.time.get_ticks()
		self.base_image = base_image # keep a original copy for tranformation
		self.image = self.base_image.copy()
		
		self.rect = self.image.get_rect()
		self.x, self.y = self.rect.center = (0,0)
		
		# self.angle = math.pi
		
		self.flightList = [
			( curves.quadratic(P(0,0),P(200,500), P(400,30), 400), 2000.0, False ),
			( curves.bezier2(P(400,30), P(0,600)), 2000.0, False )
			# ( curves.quadratic(P(0,600),P(150,250), P(400,30), 400), 1000.0, True )
			]

		self.reset()
		# self.curve = curves.quadratic(P(0,0),P(400,500), P(800,0), 800)
		# self.flightTime = float(3000) #milliseconds

	def calculatePosition(self):
		self.lastX = self.x
		self.lastY = self.y
		self.lifeTime = pygame.time.get_ticks() - self.startTime
		t = self.lifeTime / self.flightTime
		if self.reversedCurve:
			t = 1-t
		p = self.curve(t)
		self.x, self.y = p.x, p.y

	def modulate(self, periodes, dx, dy):
		timePerPeriode = self.flightTime / periodes
		timeOfPeriode = self.lifeTime % timePerPeriode
		percentage = 1 / timePerPeriode * timeOfPeriode
		self.x += dx * math.sin(math.pi * 2 *percentage)
		self.y -= dy * math.sin(math.pi * 2 *percentage)

	def rotateAlongPath(self, fixedRotation):
		#FIXME this Implementation is murks - dont use it
		dx = self.x-self.lastX
		dy = self.y-self.lastY
		if dx == 0:
			pass
		else:
			angle = math.atan2(dy,dx)
			angleDeg = -angle * 180 / math.pi + fixedRotation
			self.image = pygame.transform.rotate(self.base_image, angleDeg)

	def rotate(self, angle):
		pass

	def update(self):
			self.calculatePosition()
			self.fly()
			self.rect.center = (self.x, self.y)
			if self.lifeTime >= self.flightTime:
				if len(self.flightList) == 0:
					self.kill()
				else:
					self.reset()

	def reset(self):
		self.startTime = pygame.time.get_ticks()
		self.curve, self.flightTime, self.reversedCurve = self.flightList.pop(0)


	''' 
	this function should be varied in the different inherited Sprite-Classes
	'''

	def fly(self):
		#self.rotateAlongPath(90)
		pass


'''
class E_Hopper(Enemy):
	def __init__(self,spawnPoint, pg, reversedCurve):
		Enemy.__init__(self,spawnPoint, pg, reversedCurve)
		self.curve = curves.quadratic(P(800,spawnPoint[1]),P(400,330), P(0,spawnPoint[1]))

	def fly(self):
		self.calculatePosition()
		self.modulate(4,20,0)

	def loadImage(self):
		self.base_image = pygame.image.load(os.path.join('../data','mine.png')).convert()
		self.base_image.set_colorkey((255,0,255))
		self.image = pygame.transform.rotate(self.base_image,0)
		self.rect = self.image.get_rect()


class E_Swisher(Enemy):
	def __init__(self,spawnPoint, pg, reversedCurve):
		Enemy.__init__(self,spawnPoint, pg, reversedCurve)
		self.curve = curves.bezier4( P(200,0), P(600,300), P(150,400), P(0,600) )
		# self.curve = curves.bezier4(P(0,0),P(700,300),P(800,600))

	def calculatePosition(self):
		self.lastX = self.x
		self.lastY = self.y
		self.lifeTime = pygame.time.get_ticks() - self.startTime
		flightTime = 3000.0
		t = self.lifeTime / self.flightTime
		p = self.curve(t)
		self.x, self.y = p.x, p.y

	def fly(self):
		self.calculatePosition()

	def loadImage(self):
		self.base_image = pygame.image.load(os.path.join('../data','enemy2.png')).convert()
		self.base_image.set_colorkey((255,0,255))
		self.image = self.base_image.subsurface((22,0,22,27))
		self.rect = self.image.get_rect()
		self.rect.width /= 2
'''
		

class Spawner():
	def __init__(self, sprite, spriteGroup):
		self.spriteGroup = spriteGroup
		self.base_sprite = sprite
		# self.reversedCurve = reversedCurve
		self.spritesSpawned = 0
		# self.pg.addSpriteGroup(self.spriteGroup)
		self.startTime = pygame.time.get_ticks()
		# self.initialDelay = initialDelay
		# self.spawnSprite()
		self.numSpritesToSpawn = 0

	def spawnSingle(self, spawnPoint):
		self.spawnPoint = spawnPoint
		self.spriteGroup.add(self.base_sprite.copy())

	def spawnMulti(self, spawnPoint, numSpritesToSpawn, interval):
		self.numSpritesToSpawn = numSpritesToSpawn
		self.spawnPoint = spawnPoint
		self.interval = interval
		self.spritesSpawned += 1
		self.spriteGroup.add(self.base_sprite.copy())
		
	def update(self):
		if self.spritesSpawned < self.numSpritesToSpawn:
			self.thisTime = pygame.time.get_ticks()
			if self.thisTime - self.lastTime >= self.interval:
				self._putSprite()

	def _putSprite(self):
		self.lastTime = pygame.time.get_ticks()
		if self.lastTime - self.startTime >= self.initialDelay:
			self.spritesSpawned += 1
			self.spriteGroup.add(self.sprite(self.spawnPoint, self.pg, self.reversedCurve))

class World():
	def __init__(self, size=(1020,600), caption="NoName"):
		pygame.mixer.pre_init(44100, -16, 2, 4096)
		pygame.init()
		self.size = size
		self.playgrounds = []
		self.screen = pygame.display.set_mode(self.size)
		
		self.background = pygame.Surface(self.size)
		self.background.fill(C_BG)
		pygame.display.set_caption(caption)
		self.clock = pygame.time.Clock()
		self.frameTime = 0
		self.running = True
		self.frames = 0
		self.music = loadMusic('track-01')
		self.music.play(-1)


		#Input Control Flags
		self.input_left = False
		self.input_right = False
		self.input_up = False
		self.input_down = False

		self.input_fire1 = False
		self.input_fire2 = False

		self.input_mouse_button_left = False
		self.input_mouse_button_right = False

		self.ready_to_fire = False

	def loadImage(self):
		return pygame.image.load(data.load('background.png')).convert()

	def mainloop(self, fps):
		# ballgroup = pygame.sprite.RenderPlain()
		# laserGroup = pygame.sprite.RenderPlain()

		# enemy_spawner = Spawner(self.playgrounds[0], Enemy, 10, 400, spriteGroup=ballgroup)
		# enemy2_spawner = Spawner(self.playgrounds[0], E_Swisher, 10, 300, spawnPoint=(0,200), spriteGroup=ballgroup, initialDelay = 1000, reversedCurve=False)
		# enemy3_spawner = Spawner(self.playgrounds[0], E_Hopper, 20, 300, spawnPoint=(0,600), spriteGroup=ballgroup, initialDelay = 1500, reversedCurve=True)
		# enemy4_spawner = Spawner(self.playgrounds[0], Enemy, 10, 400, spawnPoint=(0,800), spriteGroup=ballgroup, reversedCurve=True, initialDelay=8000)
		# player = Player((400,400), self.playgrounds[0])
		# playergrounp = pygame.sprite.RenderPlain(player)
		# self.playgrounds[0].addSpriteGroup(playergrounp)
		# self.playgrounds[0].addSpriteGroup(laserGroup)

		self.fps = fps

		while self.running:
			# enemy_spawner.update()
			# enemy2_spawner.update()
			# enemy3_spawner.update()
			# enemy4_spawner.update()
			self.frameTime = self.clock.tick(self.fps)
			self.getInput()
			self.screen.blit(self.background, (0,0))
			self.updatePlaygrounds()
			pygame.display.flip()

	def getInput(self):
		events = pygame.event.get()
		for event in events:
			if event.type == QUIT:
				self.running = False
				sys.exit(0)
			if event.type == MOUSEBUTTONDOWN:
				if event.button == 1:
					self.input_mouse_button_left = True
				if event.button == 3:
					self.input_mouse_button_right = True

			if event.type == MOUSEBUTTONUP:
				if event.button == 1:
					self.input_mouse_button_left = False
				if event.button == 3:
					self.input_mouse_button_right = False
			if event.type == KEYDOWN:
				if event.key == K_LEFT or event.key == K_a:
					self.input_left = True
				if event.key == K_RIGHT or event.key == K_d:
					self.input_right = True
				if event.key == K_UP or event.key == K_w:
					self.input_up = True
				if event.key == K_DOWN or event.key == K_s:
					self.input_down = True
			if event.type == KEYUP:
				if event.key == K_LEFT or event.key == K_a:
					self.input_left = False
				if event.key == K_RIGHT or event.key == K_d:
					self.input_right = False
				if event.key == K_UP or event.key == K_w:
					self.input_up = False
				if event.key == K_DOWN or event.key == K_s:
					self.input_down = False



	def broadcast(self, eventType):
		print eventType

	def updatePlaygrounds(self):
		for pg in self.playgrounds:
			if pg.is_active:
				pg.update()

	def addPlayground(self, level):
		self.playgrounds.append(level)



class Playground():
	'''	
	this defines an area where stuff can happen
	it sits inside the world (the engine)
	levels can be loaded from here
	'''
	def __init__(self, size=(800, 600), world=None, pos=(0,0)):
		self.size = size
		self.pos = pos
		self.world = world
		self.background = pygame.Surface(self.size)
		self.background.fill(C_SPACE)
		self.spriteGroups = []
		self.is_active = True

	def update(self):
		self.drawBackground()
		self.updateSprites()
		self.drawSprites()

	def drawBackground(self):
		self.world.screen.blit(self.background, self.pos)

	def updateSprites(self):
		for spriteGroup in self.spriteGroups:
			spriteGroup.update()

	def drawSprites(self):
		for spriteGroup in self.spriteGroups:
			spriteGroup.draw(self.world.screen)

	def addSpriteGroup(self, spriteGroup):
		self.spriteGroups.append(spriteGroup)

	def addNamedSpriteGroup(self, spriteGroupName, spriteGroupType):
		eval('self.%s = pygame.sprite.%s()'%(spriteGroupName, spriteGroupType))
		self.spriteGroups.append(eval('self.%'%spriteGroupName))

	def setActive(self, status=True):
		self.is_active = status

	def loadImage(self, filename):
		return pygame.image.load(data.load(filename)).convert()


class ActionPG(Playground):
	def __init__(self, size, world, pos):
		Playground.__init__(self, size, world, pos)

		self.playerGroup = pygame.sprite.RenderPlain(Player((400,400), self))
		self.addSpriteGroup(self.playerGroup)

		# sprite = loadImage('enemy3.png', IMG_COLORKEY)
		# self.playerGroup.add(Enemy(self, sprite))

		self.laserGroup = pygame.sprite.RenderPlain()
		self.addSpriteGroup(self.laserGroup)
		# self.laser_spawner = Spawner(Laser,1)


class HoverFiring (pygame.sprite.Sprite):
	def __init__(self, pg):
		pygame.sprite.Sprite.__init__(self)
		self.pg = pg
		self.image = loadImage('firing_area_over.png')
		self.rect = self.image.get_rect()
		self.rect.topleft = (17+800,429)

	def update(self):
		pass



class ControlPG(Playground):
	def __init__(self, size, world, pos):
		Playground.__init__(self, size, world, pos)
		self.background = loadImage('control-bg.png', IMG_COLORKEY)
		
		self.hover_sprites = pygame.sprite.GroupSingle()
		self.addSpriteGroup(self.hover_sprites)

		mousePointer = pygame.sprite.RenderPlain(MousePointer(self))
		self.addSpriteGroup(mousePointer)

		self.firing_area_sprite = HoverFiring(self)

	def checkMouseOver(self):
		x,y = pygame.mouse.get_pos()
		
		self.hover_sprites.empty()
		
		if self.firing_area_sprite.rect.collidepoint((x,y)):
			self.hover_sprites.add(self.firing_area_sprite)
			self.world.ready_to_fire = True
		else:
			self.world.ready_to_fire = False
		
	def update(self):
		self.drawBackground()
		self.checkMouseOver()
		self.updateSprites()
		self.drawSprites()

class Level1():
	def __init__(self):
		song = loadMusic('track-01')
		song.play()


if __name__ == '__main__':
	game = World(caption = "MyFirstGame")
	pg1 = ActionPG((800,600), game, (0,0))
	pg2 = ControlPG((220,600), game, (800,0))
	game.addPlayground(pg1)
	game.addPlayground(pg2)
	game.mainloop(0)
